"""Names of the parser libraries."""

__all__ = ["LXML", "HTMLPARSER", "HTML5LIB", "LXML_XML", "LXML_HTML", "INIT"]

LXML = "lxml"
HTMLPARSER = "html.parser"
HTML5LIB = "html5lib"

# Alternate names for the LXML parsers
LXML_XML = "lxml-xml"
LXML_HTML = "lxml-html"

# Unknown
INIT = "[Unknown tree builder]"
