"""Some useful features for a TreeBuilder to have."""

__all__ = ["FAST", "PERMISSIVE", "STRICT", "XML", "HTML", "HTML_5"]

FAST = "fast"
PERMISSIVE = "permissive"
STRICT = "strict"
XML = "xml"
HTML = "html"
HTML_5 = "html5"
