from .attrs import AttrList
from .element_and_text import Element, TextNode
from .html5_support import HTML5TreeBuilder
from .main import TreeBuilderForHtml5lib

__all__ = [
    "HTML5TreeBuilder",
    "TreeBuilderForHtml5lib",
    "AttrList",
    "Element",
    "TextNode",
]
